import { AggregateID, AggregateRoot } from '@mobicoop/ddd-library';
import { v4 } from 'uuid';
import { AdProps, CreateAdProps, Status } from './ad.types';
import { AdCreatedDomainEvent } from './events/ad-created.domain-event';
import { AdDeletedDomainEvent } from './events/ad-delete.domain-event';
import { AdInvalidatedDomainEvent } from './events/ad-invalidated.domain-event';
import { AdSuspendedDomainEvent } from './events/ad-suspended.domain-event';
import { AdValidatedDomainEvent } from './events/ad-validated.domain-event';
import { AdUpdatedDomainEvent } from './events/ad.domain-event';
import { ScheduleItemProps } from './value-objects/schedule-item.value-object';
import { WaypointProps } from './value-objects/waypoint.value-object';

export class AdEntity extends AggregateRoot<AdProps> {
  protected readonly _id: AggregateID;

  static create = (create: CreateAdProps): AdEntity => {
    const id = v4();
    const props: AdProps = { ...create, status: Status.PENDING };
    const ad = new AdEntity({ id, props });
    ad.addEvent(
      new AdCreatedDomainEvent({
        metadata: {
          correlationId: id,
          timestamp: Date.now(),
        },
        aggregateId: id,
        userId: props.userId,
        driver: props.driver,
        passenger: props.passenger,
        frequency: props.frequency,
        fromDate: props.fromDate,
        toDate: props.toDate,
        schedule: props.schedule.map((day: ScheduleItemProps) => ({
          day: day.day,
          time: day.time,
          margin: day.margin,
        })),
        seatsProposed: props.seatsProposed,
        seatsRequested: props.seatsRequested,
        strict: props.strict,
        waypoints: props.waypoints.map((waypoint: WaypointProps) => ({
          position: waypoint.position,
          name: waypoint.address.name,
          houseNumber: waypoint.address.houseNumber,
          street: waypoint.address.street,
          postalCode: waypoint.address.postalCode,
          locality: waypoint.address.locality,
          country: waypoint.address.country,
          lon: waypoint.address.coordinates.lon,
          lat: waypoint.address.coordinates.lat,
        })),
        comment: props.comment,
      }),
    );
    return ad;
  };

  valid = (): AdEntity => {
    this.props.status = Status.VALID;
    this.addEvent(
      new AdValidatedDomainEvent({
        metadata: {
          correlationId: this.id,
          timestamp: Date.now(),
        },
        aggregateId: this.id,
      }),
    );
    return this;
  };

  invalid = (): AdEntity => {
    this.props.status = Status.INVALID;
    this.addEvent(
      new AdInvalidatedDomainEvent({
        metadata: {
          correlationId: this.id,
          timestamp: Date.now(),
        },
        aggregateId: this.id,
      }),
    );
    return this;
  };

  suspend = (): AdEntity => {
    this.props.status = Status.SUSPENDED;
    this.addEvent(
      new AdSuspendedDomainEvent({
        metadata: {
          correlationId: this.id,
          timestamp: Date.now(),
        },
        aggregateId: this.id,
      }),
    );
    return this;
  };

  update = (newProps: CreateAdProps): AdEntity => {
    this.props.driver = newProps.driver;
    this.props.passenger = newProps.passenger;
    this.props.frequency = newProps.frequency;
    this.props.fromDate = newProps.fromDate;
    this.props.toDate = newProps.toDate;
    this.props.seatsProposed = newProps.seatsProposed;
    this.props.seatsRequested = newProps.seatsRequested;
    this.props.strict = newProps.strict;
    this.props.comment = newProps.comment;
    this.props.schedule = newProps.schedule.map((item) => ({ ...item }));
    this.props.waypoints = newProps.waypoints.map((wp) => ({ ...wp }));
    //The ad goes back to pending status until it is validated again
    this.props.status = Status.PENDING;
    this.addEvent(new AdUpdatedDomainEvent(this));
    this.validate();
    return this;
  };

  delete(): void {
    this.addEvent(
      new AdDeletedDomainEvent({
        aggregateId: this.id,
      }),
    );
  }

  validate(): void {
    // entity business rules validation to protect it's invariant before saving entity to a database
  }
}
